#include	"llparam.h"
#include	"llglobal.h"
#include	<stdio.h>

/*+
 *	MODIFICATION	1
 *
 *	Purpose:	To change the representation of the parse table
 *			to be int [] rather than char [].  This removes
 *			the old limit of 255 on the number of productions
 *			in the grammar.
 *
 *	Author:		D. Bocking
 *
 *	Date:		14 December, 1982
-*/

int llst [STACKSIZE];	/* parse stack. */

int llval [VSIZE];	/* place to save attributes from llscan() */
int llvalx, *llvalp;	/* index and pointer into llval[] */

int llinhx, *llinhp;	/* index and pointer for inherited attributes */
int llsynx, *llsynp;	/* index and pointer for synthesized attributes */

int llbit;		/* indicates function of llsyn() */
			/* Either save terminal attributes or synthesize
			   attributes directly into the stack */

int	lldebug	=	0;
int	llline	=	1;


/*
 *	Main parsing algorithm for the compiler.
 */

llparse()

{
	register int *top, ssym, *rp;
	int insym, accept, *true_top, i, j, k;
	int *s, *t;
	int	*lldoact (), *lldopush (), *lldopop ();

	accept = 1;

	/*
	 *	Initialize the parse stack.
	 */
        
	top = &llst [MAXCONTEXT - 1];
	*top++ = 0;		/* lleof */
	*top = llstart;		/* Grammar start symbol */
	true_top = top;

	/*
	 *	Scan the first input symbol.
	 */

	insym = lldoscan (llval);

	for(;;)
	{
		if (lldebug & DBGSTK1) {   
			fprintf (stderr, "Primary parse stack:\n\n");
			for (rp = top; rp > llst; ) {
				fprintf (stderr, "\t%s\n",
				symbols [*rp]);

				if (*rp == 0) {
					break;
				}

				rp -= (llinhs [*rp] + llsyns [*rp] + 1);
			}
			fprintf (stderr, "\n"); 
		}

		/*
		 *	Examine symbol on top of parse stack.
		 */

		ssym = *top;

		if (lldebug & DBGSTP1) {
			fprintf (stderr, "Top of stack: %s\n", symbols [ssym]);
			fprintf (stderr, "Current input: %s\n",
			symbols [insym]);
		}

		/*
		 *	Top of stack symbol is an action routine.
		 */

		if(ssym >= llfa)
		{
			/*
			 *	Execute the action routine on the
			 *	top of the stack.
			 */

			top = lldoact (top);
			continue;
		}

		/*
		 *	Top of stack symbol is a non-terminal.  Using
		 *	that and the current input symbol, look in the
		 *	parse table for a production to push on the stack.
		 */

		if(ssym >= llfnt
		&& llchkpush (top, insym) != 0)
		{
			/*
			 *	If the parse table contains a new production
			 *	then push it on the stack in place of the
			 *	current top of stack non-terminal.
			 */

			i = llpt [(ssym - llfnt) * llfnt + insym];
			top = lldopush (top, i);

			if(llprod[i] < llprod[i + 1]) /* nonempty rhs. */
				true_top = top;
			continue;
		}

		/*
		 *	Top of stack is a terminal symbol which matches
		 *	the current input symbol ... If the input symbol is
		 *	lleof, then accept the parse.  Otherwise, pop
		 *	the stack, propogate the input symbol's attributes
		 *	into the stack, and then scan a new symbol.
		 */

		if(ssym < llfnt && insym == ssym)
		{
			if (ssym == 0) {
				return (accept);
			}

			/*
			 *	Pop the current top of stack non-terminal
			 *	symbol.
			 */

			top = lldopop (top);

			/*
			 *	Scan the next input symbol.
			 */

			insym = lldoscan (llval);

			true_top = top;
			continue;
		}

		/*
		 *	None of the above cases worked, so we have
		 *	a syntax error on our hands.  Try error recovery.
		 */

		accept = llrecover (top, insym, llval);
		if (accept == 0) {
			return (accept);
		}

		insym = lldoscan (llval);
	}
}

/*
 *	The following routine is used to execute an action routine
 *	which is on top of the parse stack.  The action routine stack
 *	item and all of its inherited and synthesized attributes
 *	are removed from the stack and the new top of stack pointer
 *	is returned to the caller.
 *
 *		Stack			Becomes
 *
 *		|        |		|        |
 *		+--------+		+--------+
 *	 top ---> action |		|        |
 *		+--------+		+--------+
 *		| inh a  |    llinhp --->        |
 *		+--------+		+--------+
 *		| inh b  |		|        |
 *		+--------+		+--------+
 *		| syn c -+--+ llsynp --->        |
 *		+--------+  |		+--------+
 *		|        |  |	 top --->        |
 *		+--------+  |		+--------+
 *		| inh c  <--+		| val c  |
 *		+--------+		+--------+
 *		|        |		|        |
 */

int	*
lldoact (top)

int	*top;

{
	int	ssym;
	/*
	 *	Pop action routine symbol from top of
	 *	stack, calculate number of inherited
	 *	and synthesized attributes and then
	 *	pop them too.  (Note that the inheritance
	 *	and synthesis pointers are initialized so
	 *	that they point directly at the stack
	 *	area just popped.
	 */

	if (lldebug & DBGSTP1) {
		fprintf (stderr, "Execute action %s\n", symbols [*top]);
	}

	ssym = *top;
	llinhp= --top;

	llinhx= llinhs[ssym];
	llsynx= llsyns[ssym];
	llsynp= llinhp - llinhx;

	top= llsynp - llsynx;

	/*
	 *	Call action routine.  It will call llinh
	 *	and llsyn which alter the llinhx and llsynx
	 *	counters tested below.
	 */

	(*funct[ssym-llfa])();
	if (llinhx)
	{
		fprintf (stderr, "COMPILER ERROR: '%s' ", symbols[ssym]);
		fprintf (stderr, "inherits wrong number of attributes.\n");
		exit(1);
	}

	if (llsynx)
	{
		fprintf (stderr, "COMPILER ERROR: '%s' ", symbols[ssym]);
		fprintf (stderr, "synthesizes wrong number of attributes.\n");
		exit(1);
	}

	return (top);
}

/*
 *	The following routine is called when a non-terminal is on
 *	the top of the stack and a production is going to be applied.
 *	It checks that:  There is a production specified in the parse
 *	table and that such a production is NOT the first of a series
 *	of null production applications which will result in a error.
 *	This enables the parser to detect an error on the FIRST appearance
 *	of a syntax error, BEFORE any null productions are applied and
 *	any associated action routines are invoked.  This early detection
 *	of syntax errors is called the Immediate Error Detection property
 *	and is required to take full advantage of the FMQ error correction
 *	mechanism.
 */

llchkpush (stktop, insym)

int	*stktop;
int	insym;

{
	int	stksym, oktopush, newprod, tokenlen;

	oktopush = 0;
	stksym = *stktop;

	newprod = llpt [(stksym - llfnt) * llfnt + insym];
	if (newprod > 0) {
		tokenlen = llrhstokens (stksym, newprod);
		if (tokenlen == 0) {
			oktopush = llpredict (stktop, insym);
		} else {
			oktopush = 1;
		}
	}

	return (oktopush);
}

/*
 *	Predict whether or not the application of the current production
 *	with a null right hand side will eventually result in an
 *	error situation.
 */

llpredict (curtop, insym)

int	*curtop;
int	insym;

{
	register	int	toptoken, newprod;

	do {
		toptoken = *curtop;

		if (toptoken < llfnt) {
			return (toptoken == insym ? 1 : 0);

		} else if (toptoken < llfa) {
			newprod = llpt [(toptoken - llfnt) * llfnt + insym];
			if (newprod == 0) {
				return (0);

			} else if (llrhstokens (toptoken, newprod) != 0) {
				return (1);

			} else {
				return (llpredict (curtop - llinhs [toptoken]
				- llsyns [toptoken] - 1, insym));
			}
		}

		curtop -= llinhs [toptoken] + llsyns [toptoken] + 1;
	} while (toptoken != 0);

	return (0);
}

llrhstokens (lhs, i)

int	lhs;
int	i;

{
	register	int	*rptr, rlen, rtokens, token;

	rlen = llrhssize (i) - llinhs [lhs] - llsyns [lhs];
	rptr = llprod [i - 1];
	rptr += llinhs [lhs] + llsyns [lhs];

	rtokens = 0;
	while (rlen > 0) {
		token = *rptr++;
		if (token < llfa) {
			rtokens++;
		}

		rptr += llinhs [token] + llsyns [token];
		rlen -= llinhs [token] + llsyns [token] + 1;
	}

	return (rtokens);
}

llrhssize (i)

int	i;

{
	return (llprod [i] - llprod [i - 1]);
}

/*
 *	The following routine is called when a non-terminal on top
 *	of the stack is to be popped and replaced by the right hand
 *	side of another production.  The non-terminal is first popped,
 *	its inherited and synthesized attributes are saved, the new
 *	production is pushed onto the stack, the replaced non-terminal's
 *	inherited attributes are propogated into the new production, and
 *	finally, each of the replaced non-terminal's synthesized attributes
 *	is adjusted so that the point of synthesis of the attribute
 *	in the new production's right hand side connects to the point
 *	of inheritance in the rest of the original production's right
 *	hand side.  (I would draw a picture of this, but it is HARD!)
 */

int	*
lldopush (top, i)

int	*top;
int	i;

{
	int	j, *rp, *thread, *stkptr, ssym;
	int	llact [VSIZE], llactx, *llactp;

	if (lldebug & DBGSTP1) {
		fprintf (stderr, "Expand symbol %s with production %d\n",
		symbols [*top], i);
	}
	/*
	 *	Pop non terminal symbol from top of
	 *	stack, and save its inherited and synthesized
	 *	attributes.
	 */

	ssym = *top;
	top--;

	llactp= llact;
	j= llinhs[ssym];
	while(j--)
		*llactp++= *top--;

	j= llsyns[ssym];
	while(j--)
	{
		if(*top == 0)
			*llactp++= 0;
		else
			/*
			 *	This saves the ABSOLUTE stack address,
			 *	and not a relative one.  This is turned
			 *	back to relative stack pointer below.
			 */

			*llactp++ = (int) (&top[*top]);
		top--;
	}

	/*
	 *	Push on the new production.
	 */

	rp = llprod[i];
	i--;

	for(rp--; rp >= llprod[i] ; rp--)
	{
		if(++top >= &llst [STACKSIZE])
		{
			fprintf (stderr, "COMPILER ERROR:  %s\n",
			"parse stack overflow");
			exit(1);
		}
		*top = *rp;
	}

	/*
	 *	Propogate the saved inherited attributes of the replaced
	 *	nonterminal symbol into the newly pushed
	 *	production.
	 */

	llactp= llact;
	j= llinhs[ssym];
	while(j--)
	{
		thread = &top[*top];
		top--;
		llpropogate (thread, *llactp++);
	}

	/*
	 *	Adjust the left-hand side's synthesized attributes
	 *	so that they point from the place of synthesis in
	 *	the right-hand side of the new production to the
	 *	place of inheritance in the rest of the right-hand
	 *	side of the previous production.  When the attribute
	 *	is actually synthesized, it will be propogated into
	 *	the proper spot in the previous production.
	 */

	j= llsyns[ssym];
	while(j--)
	{
		thread = &top[*top];
		top--;
		stkptr = (int *) (*llactp++);
		if(stkptr == 0)
			*thread = 0;
		else
			/*
			 *	This changes the absolute stack address
			 *	saved earlier back into a RELATIVE stack
			 *	which will be used when the attribute is
			 *	actually synthesized and propogated into
			 *	the stack.
			 */

			*thread = stkptr - thread;
	}

	return (top);
}

/*
 *	The following routine is used to pop a terminal symbol
 *	off the top of the stack when it matches the current
 *	input symbol.  The symbol is popped and its synthesized
 *	attributes are propogated down the stack into the
 *	remainder of the production.  Relatively straight
 *	forward ... no picture is necessary.
 */

int	*
lldopop (top)

int	*top;

{
	int	i, ssym, *thread;

	if (lldebug & DBGSTP1) {
		fprintf (stderr, "Pop terminal symbol %s\n", symbols [*top]);
	}

	/*
	 *	Pop the stack.
	 */

	ssym = *top;
	top--;

	/*
	 *	Propogate the terminal symbol's
	 *	synthesized attributes into the parse
	 *	stack.
	 */

	llvalp= llval;
	i= llsyns[ssym];

	while(i--)
	{
		thread = &top[*top];
		top--;
		llpropogate (thread, *llvalp++);
	}

	/*
	 *	Having done all that, look for any and all
	 *	action routines which immediately follow
	 *	the popped terminal symbol in the production
	 *	and execute them BEFORE scanning the next
	 *	input symbol.  This is done so that when the
	 *	language is an interactive one, actions
	 *	associated with the matched terminal are not
	 *	delayed by the process of getting the next
	 *	terminal.
	 */

	while((ssym = *top) >= llfa)
	{
		top = lldoact (top);
	}

	return (top);
}

/*
 *	This routine is used to obtain the next input symbol from
 *	the scanner.  It does the housework required to get the
 *	synthesized attributes for the terminal symbol in the right
 *	place and then checks that the number of attributes is correct.
 */
lldoscan (synptr)

int	*synptr;

{
	int	symbol, i;

	llbit = 1;	/* Tell llsyn to synthesize terminal attributes */
	llvalx = VSIZE;
	llvalp = synptr;

	symbol = llscn2 ();
	llbit = 0;

	/*
	 *	Check to make sure that the correct number of attributes
	 *	were synthesized.
	 */

	i = llsyns [symbol];
	if (i != VSIZE - llvalx) {
		fprintf (stderr, "COMPILER ERROR: llscan (%s) ",
		symbols [symbol]);
		fprintf (stderr, "synthesizes wrong number of attributes.\n");
		exit (1);
	}

	return (symbol);
}

/*
 *	This routine is used to obtain default synthesized attributes
 *	for terminal symbols inserted by the error recovery
 *	mechanism.  It does the housework required to get the
 *	synthesized attributes for the terminal symbol in the right
 *	place and then checks that the number of attributes is correct.
 */
lldogensyn (token, synptr)

int	token;
int	*synptr;

{
	int	i;

	llbit = 1;	/* Tell llsyn to synthesize terminal attributes */
	llvalx = VSIZE;
	llvalp = synptr;

	llgensyn (token);
	llbit = 0;

	/*
	 *	Check to make sure that the correct number of attributes
	 *	were synthesized.
	 */

	i = llsyns [token];
	if (i != VSIZE - llvalx) {
		fprintf (stderr, "COMPILER ERROR: llgensyn (%s) ",
		symbols [token]);
		fprintf (stderr, "synthesizes wrong number of attributes.\n");
		exit (1);
	}
}

/*
 *	The following routine is called by the scanner and action routines
 *	to synthesize attributes.
 */

llsyn(value)

int value;

{
	int	*thread;

	/*
	 *	If value is synthesized by scanner, just save it
	 *	for later propogation into a production when the
	 *	new terminal is matched to a terminal in the production.
	 *	Otherwise, the value is synthesized by an action
	 *	routine and must be propogated immediately.
	 */

	if (llbit)
	{
		if (llvalx-- > 0)
			*llvalp++= value;
	}
	else
	{
		if (llsynx-- > 0)
		{
			thread = llsynp--;
			llpropogate (thread, value);
		}
	}
}

/*
 *	This routine is used by action routines to inherite attributes.
 */

llinh()
{
	if (llinhx-- > 0)
		return(*llinhp--);

	fprintf (stderr, "COMPILER ERROR: llinh() misused\n");
	return(0);
}

/*
 *	The following routine is called to propogate a synthesized
 *	value along a thread of locations in the parse stack where
 *	the value is inherited.
 *
 *		Stack			Becomes
 *
 *		|        |		|        |
 *		+--------+		+--------+
 *	thread -> offst -+--+		| value  |
 *		+--------+  |		+--------+
 *		|        |  |		|        |
 *		+--------+  |		+--------+
 *		|        |  |		|        |
 *		+--------+  |		+--------+
 *	     +--+- offst <--+		| value  |
 *	     |  +--------+		+--------+
 *	     |  |        |		|        |
 *	     |  +--------+		+--------+
 *	     +-->   0    |		| value  |
 *		+--------+		+--------+
 *		|        |		|        |
 */

llpropogate (thread, value)

int	*thread;
int	value;

{
	int	*stkptr;

	while (*thread != 0) {
		stkptr = &thread [*thread];
		*thread = value;
		thread = stkptr;
	}

	*thread = value;
}
