/*+
 *	Module:		llprod.c
 *
 *	Purpose:	Defines routines which make entries in the
 *			production table.
 *
 *	Invocation:	mkproduction (lhs, rhs);
 *
 *			foreachprod (routine, param);
 *
 *			prproductions ();
 *
 *	Parameters:
 *
 *	  Input:	lhs - The address of a symbol table entry
 *			for the left hand side non-terminal symbol.
 *
 *			rhs - The address of the first right hand
 *			side token symbol.
 *
 *			routine - The address of a caller supplied
 *			subroutine to be called by foreachprod for
 *			each production in the table.  The routine
 *			is called as follows:
 *
 *				routine (n, lhs, rhs, rhslen, param);
 *
 *			param - A user supplied parameter to be passed
 *			to routine by foreachprod.  Extensive liberties
 *			are taken with the type of this parameter!
 *
 *	  Output:	None.
 *
 *	Author:		D. Bocking
 *
 *	Date:		30 October, 1984
-*/

#include	"llprod.h"
#include	"llparam.h"
#include	<stdio.h>
#include	"llglobal.h"

/*
 *	Storage for the production table, along with a pointer
 *	to the next available slot.
 */

struct	production	productions [MAXPROD];
struct	production	*nextprod = productions;

mkproduction (lhs, rhs)

struct	symtab	*lhs;
int	*rhs;

{
	if (nextprod >= &productions [MAXPROD]) {
		fprintf (stderr, "%d: Production table overflow!\n", line);
		exit (1);
	}

	nextprod -> p_lhs = lhs;
	nextprod -> p_rhs = rhs;
	nextprod++;
}

/*
 *	The following routine scans the production table and invokes
 *	a caller supplied subroutine for each entry.
 */

foreachprod (routine, param)

int	(*routine) ();
int	param;

{
	register	int	i;
	register	struct	production	*pp;

	i = 0;
	for (pp = productions; pp < nextprod - 1; pp++) {
		(*routine) (i, pp -> p_lhs, pp -> p_rhs, (pp + 1) -> p_rhs -
		pp -> p_rhs, param);
		i++;
	}
}

/*
 *	The following routine prints out the production table.
 */

prproductions ()

{
	register	struct	production	*pp;
	int	mkrhsindex ();

	fprintf (tables, "\n/* The Production Table */\n\n");
	fprintf (tables, "int *llprod [] = {\n");

	for (pp = productions; pp -> p_lhs != (struct symtab *) (-1); pp++) {
		fprintf (tables, "\t&llrhs [%d],\n", mkrhsindex (pp -> p_rhs));
	}

	fprintf (tables, "\t&llrhs [%d]\n};\n", mkrhsindex (pp -> p_rhs));
}
