/*+
 *	Module:		llprtables.c
 *
 *	Purpose:	Provides subroutines which print out various
 *			portions of the tables generated by llgen.
 *
 *	Invocation:	prgrammar ();
 *
 *	Parameters:
 *
 *	  Input:	<input parameters>
 *
 *	  Output:	<output parameters>
 *
 *	Author:		D. Bocking
 *
 *	Date:		21 November, 1984
-*/

#include	"llsymtab.h"
#include	"llprod.h"
#include	"llparam.h"
#include	<stdio.h>
#include	"llglobal.h"

prgrammar ()

{
	int	i, symcount [3], countsym (), numsym (), prinhs (), prsyns ();
	struct	symtab	*getsymaddr ();

	sortsymtab ();

	for (i = 0; i < 3; i++) {
		symcount [i] = 0;
	};
	foreachsym (countsym, symcount);
	fnt = symcount [0];
	fa = fnt + symcount [1];
	fu = fa + symcount [2];

	foreachsym (numsym, 0);

	fprintf (tables,
	"/* Attributed Translation System Parsing Tables */\n\n");

	/*
	 *	Print out symbolic constants for the terminal
	 *	symbols.
	 */

	fprintf (tables, "/* Symbolic Grammar Token Definitions */\n\n");
	for (i = 0; i < fnt; i++) {
		fprintf (tables, "#define\t%s\t%d\n",
		getsymaddr (i) -> sname, i);
	}

	/*
	 *	Print out the symbol name array for all grammar symbols.
	 */

	fprintf (tables, "\n/* The Grammar Symbol Name Table */\n\n");
	fprintf (tables, "char *symbols[] = {\n");
	for (i = 0; i < fu; i++) {
		fprintf (tables, "\t\"%s\",\n", getsymaddr (i) -> sname);
	}
	fprintf(tables, "};\n");

	/*
	 *	Print out some keys to the grammar tables.
	 */

	fprintf (tables, "\n/* Keys to the Grammar */\n\n");
	fprintf (tables,
	"int\tllfnt = %d;\t\t/* First Non Terminal Symbol */\n", fnt);
	fprintf (tables, "int\tllfa  = %d;\t\t/* First Action Symbol */\n",
	fa);
	fprintf (tables, "int\tllfu  = %d;\t\t/* First Unused Symbol */\n",
	fu);
	fprintf (tables, "int\tllstart = %d;\t\t/* Grammar Start Symbol */\n",
	startsymbol -> snum);


	/*
	 *	Print out the llinhs table.
	 */

	fprintf (tables,
	"\n/* Inherited and Synthesized Attribute Tables */\n\n");
	fprintf (tables, "char llinhs [] = {\n\t");
	foreachsym (prinhs, 10);
	fprintf (tables, "\n};\n\n");

	/*
	 *	Print out the llsyns table.
	 */

	fprintf (tables, "char llsyns [] = {\n\t");
	foreachsym (prsyns, 10);
	fprintf (tables, "\n};\n");

	/*
	 *	Print out the action routine definitions and the array
	 *	containing the address of each one.  When no action
	 *	symbols have been defined in the grammar, generate a
	 *	dummy entry, since C doesn't seem to like initializing
	 *	an array of addresses of routines with a null list of
	 *	entries.
	 */

	fprintf (tables,
	"\n/* Action Routine Definitions and Addresses */\n\n");
	if (fa == fu) {
		fprintf (tables, "_lldummy()\n\n{\n}\n\n");
	} else {
		for (i = fa; i < fu; i++) {
			fprintf (tables, "int\t%s();\n",
			getsymaddr (i) -> sname);
		}
	}

	fprintf(tables, "int (*funct[])() = {\n");
	if (fa == fu) {
		fprintf (tables, "\t_lldummy,\n");
	} else {
		for (i = fa; i < fu; i++) {
			fprintf (tables, "\t%s,\n", getsymaddr (i) -> sname);
		}
	}
	fprintf(tables, "};\n");

	/*
	 *	Check table size limitations prior to generating the
	 *	production and right hand side tables.
	 */

	if (fu > MAXSIZE) {
		fprintf (stderr, "Too many symbols in grammar!\n");
		exit (1);
	}

	if ((symcount [0] * symcount [1]) > MAXTABLE) {
		fprintf (stderr, "Parse table too large!\n");
		exit (1);
	}

	/*
	 *	Generate the table of right hand sides.
	 */

	prrhs ();

	/*
	 *	Generate the production table.
	 */

	prproductions ();
}

/*
 *	The following routine counts the number of symbols of each
 *	type in the symbol table.  The number of each type of symbol
 *	is returned in the appropriate entry in an array passed as
 *	a parameter by foreachsym.
 */

countsym (n, sp, count)

int	n;
struct	symtab	*sp;
int	count [];

{
	register	int	i;

	switch (sp -> stype) {

	case SYMRHS:
		i = 0;
		break;

	case SYMLHS:
		i = 1;
		break;

	case SYMACT:
		i = 2;
		break;

	default:
		fprintf (stderr, "%d: Invalid symbol type -- %s\n", line,
		sp -> sname);
		exit (1);
	}

	count [i]++;
}

/*
 *	The following routine just assigns the snum field of each
 *	of the symbols in the symbol table.
 */

numsym (n, sp, p)

int	n;
struct	symtab	*sp;
int	p;

{
	sp -> snum = n;
}

/*
 *	The following routine prints out each of the entries in the
 *	llinhs table generated by llgen.
 */

prinhs (n, sp, p)

int	n;
struct	symtab	*sp;
int	p;

{
	fprintf (tables, "%d,", sp -> inherit);
	if (n % p == p - 1) {
		fprintf (tables, "\n\t");
	}
}

/*
 *	The following routine prints out each of the entries in the
 *	llsyns table generated by llgen.
 */

prsyns (n, sp, p)

int	n;
struct	symtab	*sp;
int	p;

{
	fprintf (tables, "%d,", sp -> synth);
	if (n % p == p - 1) {
		fprintf (tables, "\n\t");
	}
}
