/*+
 *	Module:		llscan.c
 *
 *	Purpose:	To lexically analyze a grammar specification
 *			to produce tokens for use by the readgrammar
 *			routine.
 *
 *	Invocation:	token = llscan ();
 *
 *	Parameters:
 *
 *	  Input:	None.
 *
 *	  Output:	token - The numeric value corresponding to the
 *			token scanned.
 *
 *			tokenval - A global value containing additional
 *			information about the token, such as the value
 *			of a NUMBER or a pointer to an IDENT string.
 *
 *	Author:		D. Bocking
 *
 *	Date:		10 October, 1984
-*/

#include	"llscan.h"
#include	<stdio.h>
#include	"llglobal.h"
#include	<ctype.h>

union	tokenval	tokenval;

llscan ()

{
	register	char	c, *cp;
	register	int	sign, value;
	static	char	cbuf [100];

#ifdef	DEBUG
	fprintf (stderr, "llscan ()\n");
#endif
	sign = 0;	/* Assume any number will be positive */

	while (1) {
		c = fgetc (input);

		switch (c) {

		case EOF:
#ifdef	DEBUG
			fprintf (stderr, "\treturns TKEOF\n");
#endif
			return (TKEOF);

		case '$':
			while ((c = fgetc (input)) != '\n');
			/* Then fall through ... */

		case '\n':
			line++;
			/* Then fall through ... */

		case ' ':
		case '\t':
			break;

		case '>':
#ifdef	DEBUG
			fprintf (stderr, "\treturns GRTRTHN\n");
#endif
			return (GRTRTHN);

		case '<':
#ifdef	DEBUG
			fprintf (stderr, "\treturns LESSTHN\n");
#endif
			return (LESSTHN);

		case ':':
#ifdef	DEBUG
			fprintf (stderr, "\treturns COLON\n");
#endif
			return (COLON);

		case ';':
#ifdef	DEBUG
			fprintf (stderr, "\treturns SEMI\n");
#endif
			return (SEMI);

		case '%':
#ifdef	DEBUG
			fprintf (stderr, "\treturns PERCENT\n");
#endif
			return (PERCENT);

		case '|':
#ifdef	DEBUG
			fprintf (stderr, "\treturns BAR\n");
#endif
			return (BAR);

		case ',':
#ifdef	DEBUG
			fprintf (stderr, "\treturns COMMA\n");
#endif
			return (COMMA);

		case '-':
			sign = 1;

		case '+':
			/* Make sure next character is a digit! */

			if (!isdigit (c = fgetc (input))) {
				fprintf (stderr, "Bad number on line %d.  ",
				line);
				fprintf (stderr, "Skipping\n");
				break;
			}
			/* Then fall through ... */
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			/* Collect a digit string and figure out its value */

			value = c - '0';
			while (isdigit (c = fgetc (input))) {
				value *= 10;
				value += c - '0';
			}

			ungetc (c, input);
			tokenval.tk_int = value;
#ifdef	DEBUG
			fprintf (stderr, "\treturns NUMBER\n");
#endif
			return (NUMBER);

		case '@':
			c = '_';
			/* Then fall through ... */

		default:
			if (c == '_' || isalpha (c)) {
				/* Collect an identifier in the buffer */

				cp = cbuf;
				*cp++ = c;

				while (isalpha (c = fgetc (input))
				|| isdigit (c) || c == '_') {
					*cp++ = c;
				}
				*cp = '\0';
				ungetc (c, input);

				tokenval.tk_str = cbuf;
				if (cbuf [0] == '_') {
#ifdef	DEBUG
					fprintf (stderr,
					"\treturns ACTION %s\n",  cbuf);
#endif
					return (ACTION);
				} else {
#ifdef	DEBUG
					fprintf (stderr,
					"\treturns IDENT %s\n", cbuf);
#endif
					return (IDENT);
				}
			} else {
				fprintf (stderr,
				"Bad input character on line %d.  ",
				line);
				fprintf (stderr, "Skipping\n'");
			}
		}
	}
}
